==> yams-1.0.0.tar.gz: yams-1.0.0/src/memory.c <==
/* yams -- Yet Another Machine Simulator
   Copyright (C) 2002 Juha Aatrokoski, Timo Lilja, Leena Salmela,
   Teemu Takanen, Aleksi Virtanen

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.

   $Id: memory.c,v 1.31 2002/11/26 11:29:41 ttakanen Exp $
*/

/* AIX requires this to be the first thing in the file.  */

#if defined (_AIX) && !defined (__GNUC__)
#pragma alloca
#endif

#include "includes.h"
#include <assert.h>
#include <netinet/in.h>
#include <string.h>
#include "cpu.h"
#include "memory.h"
#include "simulator.h"

memory_t * memory_create(uint32_t pagesize, uint32_t num_pages) {
    memory_t *memory;

    memory               = (memory_t *) smalloc(sizeof(memory_t));
    memory->pagesize     = pagesize; 
    memory->num_pages    = num_pages;
    memory->physmem      = smalloc(num_pages * pagesize);
    memory->io_descrarea = smalloc(IO_DESCRAREA_LENGTH);
    memory->kernel_paramarea = smalloc(KERNEL_PARAMAREA_LENGTH);

    memset(memory->physmem, 0, memory->pagesize * memory->num_pages);
    memset(memory->io_descrarea, 0, IO_DESCRAREA_LENGTH);
    memset(memory->kernel_paramarea, 0, KERNEL_PARAMAREA_LENGTH);

    return memory;
}

/* does not check TLB exceptions so make sure they do not 
 * occur when calling this ;) (e.g. by doing memory load/store _before_)
 */
uint32_t phys_addr(uint32_t vaddr, cpu_t *cpu) {
    uint32_t paddr;

    if (vaddr >= 0xC0000000 || vaddr < 0x80000000) {
	/* kernel/supervisor/user mapped */
	if (tlb_translate(cpu, vaddr, &paddr, MEM_LOAD) != NoException)
	    return 0xffffffff;
    } else if (vaddr >= 0xA0000000) {
	/* kernel unmapped uncached */
	paddr = vaddr - 0xA0000000;
    } else {
	/* kernel unmapped @0x80000000 */
	paddr = vaddr - 0x80000000;
    }

    return paddr;
}

exception_t mem_read(memory_t * mem, uint32_t addr, void * buf, 
                     int size, cpu_t *cpu) {
    int ioread = 0;
    uint8_t *readfrom;
    exception_t exc;

    assert(size == 0 || size == 1 || size == 2 || size == 4);
 
    if (size != 0 && (addr & (size - 1)))
	return AddressLoad; /* alignment error */

    readfrom = (uint8_t*)mem->physmem; /* default to normal memory read */

    /* Segment translation */
    if (addr >= 0xC0000000) {
	/* 0xE0000000: kernel mapped
	 * 0xC0000000: supervisor mapped
	 * treat supervisor mode as kernel mode */
	if (!CP0_KERNEL_MODE(cpu)) return AddressLoad;
	exc = tlb_translate(cpu, addr, &addr, MEM_LOAD);
	if (exc != NoException)
	    return exc;
    } else if (addr >= 0xA0000000) { /* kernel unmapped uncached */
	if (!CP0_KERNEL_MODE(cpu)) return AddressLoad;

	/* IO area */
	if (addr >= 0xB0000000)
	    ioread = 3;

	/* IO descriptor area */
	if (addr >= IO_DESCRAREA_BASE 
	    && addr < IO_DESCRAREA_BASE + IO_DESCRAREA_LENGTH)
	    ioread = 1;

	/* Kernel boot string area */
	else if (addr >= KERNEL_PARAMAREA_BASE 
	    && addr < KERNEL_PARAMAREA_BASE + KERNEL_PARAMAREA_LENGTH)
	    ioread = 4;

	/* IO device area */
	else if (addr >= mem->io_area_base
		 && addr < mem->io_area_base + mem->io_area_length)
	    ioread = 2;

	/* "normal" area */
	else
	    addr -= 0xA0000000;

    } else if (addr >= 0x80000000) { /* kernel unmapped */
	if (!CP0_KERNEL_MODE(cpu)) return AddressLoad;
	addr -= 0x80000000;
    } else { /* user mapped */
	exc = tlb_translate(cpu, addr, &addr, MEM_LOAD);
	if (exc != NoException)
	    return exc;
    }


    /* Unused IO area read, return 0 */
    if (ioread == 3) {
	if (size != 0) memset(buf, 0, size);
	return NoException;
    }
    /* IO descriptor area read */
    if (ioread == 1) {
	addr -= IO_DESCRAREA_BASE;
	readfrom = (uint8_t*)mem->io_descrarea;
    }
    /* Kernel parameter area read */
    if (ioread == 4) {
	addr -= KERNEL_PARAMAREA_BASE;
	readfrom = (uint8_t*)mem->kernel_paramarea;
    }
    /* IO device area read */
    if (ioread == 2) {
	uint32_t word, i;
	int result;

	i = (addr - mem->io_area_base) >> 2;
	result = mem->io_device_read[i](mem->io_devices[i],
					((addr - mem->io_devices[i]->io_base) 
					 & 0xfffffffc),
					&word);
	assert(result == 0); /* Caller's fault if nonzero */

	switch(size) {
	case 4:
	    *((uint32_t*)buf) = word;
	    break;
	case 2:
	    if ((addr & 3) == 0) /* want first=high halfword */
		word >>= 16;
	    *((uint16_t*)buf) = (uint16_t)(word & 0xffff);
	    break;
	case 1:
	    word >>= 8*(3 - (addr & 3)); /* select byte */
	    *((uint8_t*)buf) = (uint8_t)(word & 0xff);
	}

	return NoException;
    }

   
    /* Out of memory bounds, normal memory */
    if (!ioread  &&  addr >= mem->num_pages * mem->pagesize)
	return BusErrorData; /* Caller checks for Instr case */

    /* normal read */
    switch(size) {
    case 1:
	*((uint8_t*)buf)
            = *(readfrom + addr);
	break;
    case 2:
	*((uint16_t*)buf)
            = ntohs(*((uint16_t*)(readfrom + addr)));
	break;
    case 4:
	*((uint32_t*)buf)
            = ntohl(*((uint32_t*)(readfrom + addr)));
    }

    return NoException;
}

exception_t mem_write(memory_t *mem, uint32_t addr, void *buf, 
                      int size, cpu_t *cpu) {
    int iowrite = 0;
    exception_t exc;

    assert(size == 0 || size == 1 || size == 2 || size == 4);

    if (size != 0 && (addr & (size - 1)))
	return AddressStore; /* alignment error */

   
    /* Segment translation */
    if (addr >= 0xC0000000) {
	/* 0xE0000000: kernel mapped
	 * 0xC0000000: supervisor mapped
	 * treat supervisor mode as kernel mode */
	if (!CP0_KERNEL_MODE(cpu)) return AddressLoad;
	exc = tlb_translate(cpu, addr, &addr, MEM_STORE);
	if (exc != NoException)
	    return exc;
    } else if (addr >= 0xA0000000) {/* kernel unmapped uncached */
	if (!CP0_KERNEL_MODE(cpu)) return AddressStore;

	/* IO area */
	if (addr >= 0xB0000000)
	    iowrite = 3;

	/* IO descriptor area */
	if (addr >= IO_DESCRAREA_BASE 
	    && addr < IO_DESCRAREA_BASE + IO_DESCRAREA_LENGTH)
	    iowrite = 1;

	/* IO device area */
	else if (addr >= mem->io_area_base
		 && addr < mem->io_area_base + mem->io_area_length)
	    iowrite = 2;

	/* "normal" area */
	else
	    addr -= 0xA0000000;

    } else if (addr >= 0x80000000) { /* kernel unmapped */
	if (!CP0_KERNEL_MODE(cpu)) return AddressStore;
	addr -= 0x80000000;
    } else { /* user mapped */
	exc = tlb_translate(cpu, addr, &addr, MEM_STORE);
	if (exc != NoException)
	    return exc;
    }


    /* Unused IO area write, do nothing */
    if (iowrite == 3) return NoException;
    /* IO descriptor area is read-only but does not cause exceptions */
    if (iowrite == 1) return NoException;
    /* Write to IO devices */
    if (iowrite == 2) {
	uint32_t word, i;
	int result;

	switch(size) {
	case 4:
	    word = *((uint32_t*)buf);
	    break;
	case 2:
	    word = (uint32_t)(*((uint16_t*)buf));
	    if ((addr & 3) == 0) /* write to first=high halfword */
		word <<= 16;
	    break;
	case 1:
	    word = (uint32_t)(*((uint8_t*)buf));
	    word <<= 8*(3 - (addr & 3)); /* write the right byte */
	}

	if (size != 0) {
	    i = (addr - mem->io_area_base) >> 2;
	    result = mem->io_device_write[i](mem->io_devices[i],
					     (addr 
					      - mem->io_devices[i]->io_base)
					     & 0xfffffffc,
					     word);
	    assert(result == 0); /* Caller's fault if nonzero */
	}

	return NoException;
    }


    /* Out of memory bounds */
    if (addr >= mem->num_pages * mem->pagesize)
	return BusErrorData; /* Caller checks for Instr case */

    /* normal write */
    switch(size) {
    case 1:
	*((uint8_t*)mem->physmem + addr)
            = *((uint8_t*)buf);
	break;
    case 2:
	*((uint16_t*)((uint8_t*)mem->physmem + addr))
            = htons(*((uint16_t*)buf));
	break;
    case 4:
	*((uint32_t*)((uint8_t*)mem->physmem + addr))
            = htonl(*((uint32_t*)buf));
    }

    return NoException;
}



int mem_dread(memory_t *mem, uint32_t addr, void *buf, int size) {

    assert(size == 1 || size == 2 || size == 4);

    /* is there need to check the alignment? */
#if 1
    if (addr & (size - 1))
	return 1; /* alignment error */
#endif   

    /* Out of memory bounds */
    if (addr > mem->num_pages * mem->pagesize - size)
	return 2;

    switch(size) {
    case 1:
	*((uint8_t*)buf)
            = *(((uint8_t*) mem->physmem) + addr);
	break;
    case 2:
	*((uint16_t*)buf)
            = ntohs(*((uint16_t*)((uint8_t*)mem->physmem + addr)));
	break;
    case 4:
	*((uint32_t*)buf)
            = ntohl(*((uint32_t*)((uint8_t*)mem->physmem + addr)));
    }

    return 0;
}

int mem_dwrite(memory_t *mem, uint32_t addr, void *buf, int size) {
    assert(size == 1 || size == 2 || size == 4);

    /* is there need to check the alignment? */
#if 1
    if (addr & (size - 1))
	return 1; /* alignment error */
#endif

    /* Out of memory bounds */
    if (addr > mem->num_pages * mem->pagesize - size)
	return 2;

    switch(size) {
    case 1:
	*((uint8_t*)mem->physmem + addr)
            = *((uint8_t*)buf);
	break;
    case 2:
	*((uint16_t*)((uint8_t*)mem->physmem + addr))
            = htons(*((uint16_t*)buf));
	break;
    case 4:
	*((uint32_t*)((uint8_t*)mem->physmem + addr))
            = htonl(*((uint32_t*)buf));
    }

    return 0;
}




int mem_store_direct(memory_t *mem, uint32_t addr, 
		     uint32_t length, void *data) {
    uint32_t size = mem->num_pages * mem->num_pages; 

    if(addr+length > size) {
	/* write over end of memory */
	return 1;
    }

    memcpy((char*)mem->physmem + addr, data, length);
    
    return 0;
} 

int mem_read_direct(memory_t *mem, uint32_t addr, 
		    uint32_t length, void *data) {
    uint32_t size = mem->num_pages *mem->num_pages; 

    if(addr+length > size) {
	/* read over end of memory */
	return 1;
    }

    memcpy(data, (char*)mem->physmem + addr, length);
    
    return 0;
}

==> yams-1.2.2.tar.gz: yams-1.2.2/src/memory.c <==
/* yams -- Yet Another Machine Simulator
   Copyright (C) 2002 Juha Aatrokoski, Timo Lilja, Leena Salmela,
   Teemu Takanen, Aleksi Virtanen

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.

   $Id: memory.c,v 1.32 2003/05/22 11:01:13 ttakanen Exp $
*/

/* AIX requires this to be the first thing in the file.  */

#if defined (_AIX) && !defined (__GNUC__)
#pragma alloca
#endif

#include "includes.h"
#include <assert.h>
#include <netinet/in.h>
#include <string.h>
#include "cpu.h"
#include "memory.h"
#include "simulator.h"

memory_t * memory_create(uint32_t pagesize, uint32_t num_pages) {
    memory_t *memory;

    memory               = (memory_t *) smalloc(sizeof(memory_t));
    memory->pagesize     = pagesize; 
    memory->num_pages    = num_pages;
    memory->physmem      = smalloc(num_pages * pagesize);
    memory->io_descrarea = smalloc(IO_DESCRAREA_LENGTH);
    memory->kernel_paramarea = smalloc(KERNEL_PARAMAREA_LENGTH);

    memset(memory->physmem, 0, memory->pagesize * memory->num_pages);
    memset(memory->io_descrarea, 0, IO_DESCRAREA_LENGTH);
    memset(memory->kernel_paramarea, 0, KERNEL_PARAMAREA_LENGTH);

    return memory;
}

/* does not check TLB exceptions so make sure they do not 
 * occur when calling this ;) (e.g. by doing memory load/store _before_)
 */
uint32_t phys_addr(uint32_t vaddr, cpu_t *cpu) {
    uint32_t paddr;

    if (vaddr >= 0xC0000000 || vaddr < 0x80000000) {
	/* kernel/supervisor/user mapped */
	if (tlb_translate(cpu, vaddr, &paddr, MEM_LOAD) != NoException)
	    return 0xffffffff;
    } else if (vaddr >= 0xA0000000) {
	/* kernel unmapped uncached */
	paddr = vaddr - 0xA0000000;
    } else {
	/* kernel unmapped @0x80000000 */
	paddr = vaddr - 0x80000000;
    }

    return paddr;
}

exception_t mem_read(memory_t * mem, uint32_t addr, void * buf, 
                     int size, cpu_t *cpu) {
    int ioread = 0;
    uint8_t *readfrom;
    exception_t exc;

    assert(size == 0 || size == 1 || size == 2 || size == 4);
 
    if (size != 0 && (addr & (size - 1)))
	return AddressLoad; /* alignment error */

    readfrom = (uint8_t*)mem->physmem; /* default to normal memory read */

    /* Segment translation */
    if (addr >= 0xC0000000) {
	/* 0xE0000000: kernel mapped
	 * 0xC0000000: supervisor mapped
	 * treat supervisor mode as kernel mode */
	if (!CP0_KERNEL_MODE(cpu)) return AddressLoad;
	exc = tlb_translate(cpu, addr, &addr, MEM_LOAD);
	if (exc != NoException)
	    return exc;
    } else if (addr >= 0xA0000000) { /* kernel unmapped uncached */
	if (!CP0_KERNEL_MODE(cpu)) return AddressLoad;

	/* IO area */
	if (addr >= 0xB0000000)
	    ioread = 3;

	/* IO descriptor area */
	if (addr >= IO_DESCRAREA_BASE 
	    && addr < IO_DESCRAREA_BASE + IO_DESCRAREA_LENGTH)
	    ioread = 1;

	/* Kernel boot string area */
	else if (addr >= KERNEL_PARAMAREA_BASE 
	    && addr < KERNEL_PARAMAREA_BASE + KERNEL_PARAMAREA_LENGTH)
	    ioread = 4;

	/* IO device area */
	else if (addr >= mem->io_area_base
		 && addr < mem->io_area_base + mem->io_area_length)
	    ioread = 2;

	/* "normal" area */
	else
	    addr -= 0xA0000000;

    } else if (addr >= 0x80000000) { /* kernel unmapped */
	if (!CP0_KERNEL_MODE(cpu)) return AddressLoad;
	addr -= 0x80000000;
    } else { /* user mapped */
	exc = tlb_translate(cpu, addr, &addr, MEM_LOAD);
	if (exc != NoException)
	    return exc;
    }


    /* Unused IO area read, return 0 */
    if (ioread == 3) {
	if (size != 0) memset(buf, 0, size);
	return NoException;
    }
    /* IO descriptor area read */
    if (ioread == 1) {
	addr -= IO_DESCRAREA_BASE;
	readfrom = (uint8_t*)mem->io_descrarea;
    }
    /* Kernel parameter area read */
    if (ioread == 4) {
	addr -= KERNEL_PARAMAREA_BASE;
	readfrom = (uint8_t*)mem->kernel_paramarea;
    }
    /* IO device area read */
    if (ioread == 2) {
	uint32_t word, i;
	int result;

	i = (addr - mem->io_area_base) >> 2;
	result = mem->io_device_read[i](mem->io_devices[i],
					((addr - mem->io_devices[i]->io_base) 
					 & 0xfffffffc),
					&word);
	assert(result == 0); /* Caller's fault if nonzero */

	switch(size) {
	case 4:
	    *((uint32_t*)buf) = word;
	    break;
	case 2:
	    if ((addr & 3) == 0) /* want first=high halfword */
		word >>= 16;
	    *((uint16_t*)buf) = (uint16_t)(word & 0xffff);
	    break;
	case 1:
	    word >>= 8*(3 - (addr & 3)); /* select byte */
	    *((uint8_t*)buf) = (uint8_t)(word & 0xff);
	}

	return NoException;
    }

   
    /* Out of memory bounds, normal memory */
    if (!ioread  &&  addr >= mem->num_pages * mem->pagesize)
	return BusErrorData; /* Caller checks for Instr case */

    /* normal read */
    switch(size) {
    case 1:
	*((uint8_t*)buf)
            = *(readfrom + addr);
	break;
    case 2:
	*((uint16_t*)buf)
            = ntohs(*((uint16_t*)(readfrom + addr)));
	break;
    case 4:
	*((uint32_t*)buf)
            = ntohl(*((uint32_t*)(readfrom + addr)));
    }

    return NoException;
}

exception_t mem_write(memory_t *mem, uint32_t addr, void *buf, 
                      int size, cpu_t *cpu) {
    int iowrite = 0;
    exception_t exc;

    assert(size == 0 || size == 1 || size == 2 || size == 4);

    if (size != 0 && (addr & (size - 1)))
	return AddressStore; /* alignment error */

   
    /* Segment translation */
    if (addr >= 0xC0000000) {
	/* 0xE0000000: kernel mapped
	 * 0xC0000000: supervisor mapped
	 * treat supervisor mode as kernel mode */
	if (!CP0_KERNEL_MODE(cpu)) return AddressLoad;
	exc = tlb_translate(cpu, addr, &addr, MEM_STORE);
	if (exc != NoException)
	    return exc;
    } else if (addr >= 0xA0000000) {/* kernel unmapped uncached */
	if (!CP0_KERNEL_MODE(cpu)) return AddressStore;

	/* IO area */
	if (addr >= 0xB0000000)
	    iowrite = 3;

	/* IO descriptor area */
	if (addr >= IO_DESCRAREA_BASE 
	    && addr < IO_DESCRAREA_BASE + IO_DESCRAREA_LENGTH)
	    iowrite = 1;

	/* IO device area */
	else if (addr >= mem->io_area_base
		 && addr < mem->io_area_base + mem->io_area_length)
	    iowrite = 2;

	/* "normal" area */
	else
	    addr -= 0xA0000000;

    } else if (addr >= 0x80000000) { /* kernel unmapped */
	if (!CP0_KERNEL_MODE(cpu)) return AddressStore;
	addr -= 0x80000000;
    } else { /* user mapped */
	exc = tlb_translate(cpu, addr, &addr, MEM_STORE);
	if (exc != NoException)
	    return exc;
    }


    /* Unused IO area write, do nothing */
    if (iowrite == 3) return NoException;
    /* IO descriptor area is read-only but does not cause exceptions */
    if (iowrite == 1) return NoException;
    /* Write to IO devices */
    if (iowrite == 2) {
	uint32_t word, i;
	int result;

	switch(size) {
	case 4:
	    word = *((uint32_t*)buf);
	    break;
	case 2:
	    word = (uint32_t)(*((uint16_t*)buf));
	    if ((addr & 3) == 0) /* write to first=high halfword */
		word <<= 16;
	    break;
	case 1:
	    word = (uint32_t)(*((uint8_t*)buf));
	    word <<= 8*(3 - (addr & 3)); /* write the right byte */
	    break;
	default:
	    fprintf(stderr, "internal error in iowrite\n");
	    exit(1);
	}

	if (size != 0) {
	    i = (addr - mem->io_area_base) >> 2;
	    result = mem->io_device_write[i](mem->io_devices[i],
					     (addr 
					      - mem->io_devices[i]->io_base)
					     & 0xfffffffc,
					     word);
	    assert(result == 0); /* Caller's fault if nonzero */
	}

	return NoException;
    }


    /* Out of memory bounds */
    if (addr >= mem->num_pages * mem->pagesize)
	return BusErrorData; /* Caller checks for Instr case */

    /* normal write */
    switch(size) {
    case 1:
	*((uint8_t*)mem->physmem + addr)
            = *((uint8_t*)buf);
	break;
    case 2:
	*((uint16_t*)((uint8_t*)mem->physmem + addr))
            = htons(*((uint16_t*)buf));
	break;
    case 4:
	*((uint32_t*)((uint8_t*)mem->physmem + addr))
            = htonl(*((uint32_t*)buf));
    }

    return NoException;
}



int mem_dread(memory_t *mem, uint32_t addr, void *buf, int size) {

    assert(size == 1 || size == 2 || size == 4);

    /* is there need to check the alignment? */
#if 1
    if (addr & (size - 1))
	return 1; /* alignment error */
#endif   

    /* Out of memory bounds */
    if (addr > mem->num_pages * mem->pagesize - size)
	return 2;

    switch(size) {
    case 1:
	*((uint8_t*)buf)
            = *(((uint8_t*) mem->physmem) + addr);
	break;
    case 2:
	*((uint16_t*)buf)
            = ntohs(*((uint16_t*)((uint8_t*)mem->physmem + addr)));
	break;
    case 4:
	*((uint32_t*)buf)
            = ntohl(*((uint32_t*)((uint8_t*)mem->physmem + addr)));
    }

    return 0;
}

int mem_dwrite(memory_t *mem, uint32_t addr, void *buf, int size) {
    assert(size == 1 || size == 2 || size == 4);

    /* is there need to check the alignment? */
#if 1
    if (addr & (size - 1))
	return 1; /* alignment error */
#endif

    /* Out of memory bounds */
    if (addr > mem->num_pages * mem->pagesize - size)
	return 2;

    switch(size) {
    case 1:
	*((uint8_t*)mem->physmem + addr)
            = *((uint8_t*)buf);
	break;
    case 2:
	*((uint16_t*)((uint8_t*)mem->physmem + addr))
            = htons(*((uint16_t*)buf));
	break;
    case 4:
	*((uint32_t*)((uint8_t*)mem->physmem + addr))
            = htonl(*((uint32_t*)buf));
    }

    return 0;
}




int mem_store_direct(memory_t *mem, uint32_t addr, 
		     uint32_t length, void *data) {
    uint32_t size = mem->num_pages * mem->pagesize; 

    if(addr+length > size) {
	/* write over end of memory */
	return 1;
    }

    memcpy((char*)mem->physmem + addr, data, length);
    
    return 0;
} 

int mem_read_direct(memory_t *mem, uint32_t addr, 
		    uint32_t length, void *data) {
    uint32_t size = mem->num_pages * mem->pagesize; 

    if(addr+length > size) {
	/* read over end of memory */
	return 1;
    }

    memcpy(data, (char*)mem->physmem + addr, length);
    
    return 0;
}
